try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps  # Python 2.3, 2.4 fallback.
import datetime, re

# Simple functions for extracting input tags from HTML - when you're testing 
# CSRF protection you need to be able to pull out the
#     <input type="hidden" name="csrf_token" value="...">
# tags so you can POST a correct value.

input_re = re.compile('<input(.*?)>')
attrib_re = re.compile(r'(\w+)=(["\'])(.*?)(\2)')

def extract_input_tag_attrs(html):
    "Returns a list of HTML attribute dictionaries for all inputs found"
    input_tags = []
    for input in input_re.findall(html):
        pairs = []
        for bits in attrib_re.findall(input):
            pairs.append((bits[0], bits[2]))
        input_tags.append(dict(pairs))
    return input_tags

def extract_input_tags(html):
    "Returns a dictionary of name => value for all inputs found"
    tag_attrs = extract_input_tag_attrs(html)
    return dict([
        (d['name'], d.get('value', ''))
        for d in tag_attrs
        if d.has_key('name')
    ])

# Decorator for temporarily monkey-patching datetime.utcnow

def fake_utcnow(fake):
    def outer(fn):
        def inner(*args, **kwargs):
            class _Object: pass
            orig_datetime = datetime.datetime
            fake_datetime = _Object()
            fake_datetime.utcnow = lambda: fake
            datetime.datetime = fake_datetime
            try:
                ret_value = fn(*args, **kwargs)
            finally: # or exception raised by fn() will not revert monkeypatch
                datetime.datetime = orig_datetime
            return ret_value
        return wraps(fn)(inner)
    return outer



from django.conf.urls.defaults import patterns
from django import forms
from django.forms.formsets import formset_factory
from django.http import HttpResponse, HttpResponseRedirect
from django_safeform import SafeForm, csrf_protect, csrf_utils, CsrfForm

class BasicForm(forms.Form):
    name = forms.CharField(max_length = 100)
SafeBasicForm = SafeForm(BasicForm)

class OtherForm(forms.Form):
    email = forms.EmailField()
SafeOtherForm = SafeForm(OtherForm)

@csrf_protect
def safe_form_view(request):
    form = SafeBasicForm(request)
    if form.is_valid():
        return HttpResponse('Valid: %s' % form.cleaned_data['name'])
    return HttpResponse("""
        <form action="." method="post">
        %s
        <p><input type="submit"></p>
        </form>
    """ % form.as_p())

@csrf_protect
def hand_rolled_view(request):
    if request.method == 'POST':
        csrf_token = request.POST.get('csrf_token', '')
        if not csrf_utils.validate_csrf_token(csrf_token, request):
            return HttpResponse('Invalid CSRF token')
        else:
            return HttpResponse('OK')
    else:
        return HttpResponse("""
        <form action="." method="post">
        <input type="text" name="name">
        <input type="hidden" name="csrf_token" value="%s">
        </form>
        """ % csrf_utils.new_csrf_token(request))

@csrf_protect
def two_forms_view(request):
    form1 = SafeBasicForm(request)
    form2 = SafeOtherForm(request)
    if form1.is_valid() and form2.is_valid():
        return HttpResponse(
            'Valid: %s, %s' % (
                form1.cleaned_data['name'], 
                form2.cleaned_data['email'],
            )
        )
    return HttpResponse("""
        <form action="." method="post">
        %s %s
        <p><input type="submit"></p>
        </form>
    """ % (form1.as_p(), form2.as_p()))

@csrf_protect
def safe_form_custom_message_view(request):
    form = SafeForm(BasicForm, invalid_message='Oh no!')(request)
    if form.is_valid():
        return HttpResponse('Valid')
    return HttpResponse(form.as_p())

@csrf_protect
def safe_form_ajax_skips_false_view(request):
    form = SafeForm(BasicForm, ajax_skips_check=False)(request)
    if form.is_valid():
        return HttpResponse('Valid')
    return HttpResponse(form.as_p())

class PersonForm(forms.Form):
    name = forms.CharField(max_length = 100)
    email = forms.EmailField()

PersonFormSet = formset_factory(PersonForm, extra=3)

@csrf_protect
def safe_formset_view(request):
    csrf_form = CsrfForm(request)
    formset = PersonFormSet()
    if request.method == 'POST':
        formset = PersonFormSet(request.POST)
        if csrf_form.is_valid() and formset.is_valid():
            return HttpResponse('Valid: %s' % ', '.join([
                '%(name)s [%(email)s]' % form.cleaned_data 
                for form in formset.forms
                if form.cleaned_data
            ]))
    return HttpResponse("""
        <form action="." method="post">
        %s %s %s
        <p><input type="submit"></p>
        </form>
    """ % (
        csrf_form,
        ''.join([f.as_p() for f in formset.forms]),
        formset.management_form
    ))

@csrf_protect
def identifier_form_view(request):
    form = SafeForm(BasicForm, identifier='identifier-form')(request)
    if form.is_valid():
        return HttpResponse('Valid: %s' % form.cleaned_data['name'])
    return HttpResponse(form.as_p())

@csrf_protect
def expire_after_60_seconds_form_view(request):
    form = SafeForm(BasicForm, expire_after=60)(request)
    if form.is_valid():
        return HttpResponse('Valid: %s' % form.cleaned_data['name'])
    return HttpResponse(form.as_p())

urlpatterns = patterns('',
    (r'^safe-basic-form/$', safe_form_view),
    (r'^two-forms/$', two_forms_view),
    (r'^safe-formset/$', safe_formset_view),
    (r'^safe-form-custom-message/$', safe_form_custom_message_view),
    (r'^safe-form-ajax-skips-false/$', safe_form_ajax_skips_false_view),    
    (r'^hand-rolled/$', hand_rolled_view),
    (r'^identifier-form/$', identifier_form_view),
    (r'^expire-after-60-seconds/$', expire_after_60_seconds_form_view),
)



from django.utils.hashcompat import sha_constructor as sha1
import random
try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps  # Python 2.3, 2.4 fallback.

def csrf_protect(view_func):
    def inner(request, *args, **kwargs):
        need_to_set_the_cookie = False
        csrf_token = request.COOKIES.get('_csrf_cookie')
        if not csrf_token:
            need_to_set_the_cookie = True
            csrf_token = sha1(str(random.random())).hexdigest()
            request._csrf_token_to_set = csrf_token
        response = view_func(request, *args, **kwargs)
        if need_to_set_the_cookie:
            response.set_cookie('_csrf_cookie', csrf_token)
        return response
    return wraps(view_func)(inner)



import time, datetime, hmac
from django.utils.hashcompat import sha_constructor as sha1
from django.conf import settings

def _csrf_token_from_request(request):
    if hasattr(request, '_csrf_token_to_set'):
        return request._csrf_token_to_set
    return request.COOKIES.get('_csrf_cookie', '')

def new_csrf_token(request, identifier='default'):
    epoch_time = int(
        time.mktime(datetime.datetime.utcnow().timetuple())
    )
    message = '%s:%s' % (identifier, epoch_time)
    secret_key = settings.SECRET_KEY + _csrf_token_from_request(request)
    sig = hmac.new(secret_key, message, sha1).hexdigest()
    return '%s:%s' % (message, sig)

not_set = object()

def validate_csrf_token(token, request, identifier='default', 
        expire_after=not_set):
    if not ':' in token:
        return False
    message, signature = token.rsplit(':', 1)
    secret_key = settings.SECRET_KEY + _csrf_token_from_request(request)
    expected_sig = hmac.new(secret_key, message, sha1).hexdigest()
    if signature != expected_sig:
        return False
    
    # Check the expiry and identifier
    token_identifier, created_at = message.split(':', 1)
    if token_identifier != identifier:
        return False
    
    if expire_after is not_set:
        expire_after = getattr(settings, 'CSRF_TOKENS_EXPIRE_AFTER', None)
    
    if expire_after is not None:
        epoch_time = int(
            time.mktime(datetime.datetime.utcnow().timetuple())
        )
        if int(created_at) + expire_after < epoch_time:
            return False
    
    return True



"""
Things to test:

1. That a form can be wrapped in a SafeForm
2. Error if cookie not present
3. Error if cookie present but altered
4. Error if csrf form value is incorrect
5. Error if csrf form value is tampered with
6. Error if csrf form value is more than CSRF_TOKEN_EXPIRES old
7. That SafeForm works with ModelForms, FormSets and so on
8. That a cookie is set by the decorator
9. That hand-rolled forms can be neatly protected as well
10. Form identifiers can be used to tie CSRF tokens to an individual form
"""

from django.test import TestCase
from django_safeform import csrf_utils
from django_safeform import test_utils
from django_safeform.forms import CSRF_INVALID_MESSAGE
import datetime

class SafeBasicFormTest(TestCase):
    urls = 'django_safeform.test_views'
    
    def test_cookie_is_set(self):
        response = self.client.get('/safe-basic-form/')
        self.assert_(response.cookies.has_key('_csrf_cookie'))
    
    def test_submission_with_correct_csrf_token_works(self):
        response = self.client.get('/safe-basic-form/')
        inputs = test_utils.extract_input_tags(response.content)
        self.assert_(inputs.has_key('csrf_token'))
        token = inputs['csrf_token']
        response2 = self.client.post('/safe-basic-form/', {
            'csrf_token': token,
            'name': 'Test',
        })
        self.assertEqual(response2.content, 'Valid: Test')
    
    def test_submission_with_bad_token_fails(self):
        response = self.client.post('/safe-basic-form/', {
            'csrf_token': 'bad-token',
            'name': 'Test',
        })
        self.assert_(CSRF_INVALID_MESSAGE in response.content)
    
    def test_invalid_message_argument_sets_custom_message(self):
        response = self.client.post('/safe-form-custom-message/', {
            'csrf_token': 'bad-token',
            'name': 'Test',
        })
        self.assert_('Oh no!' in response.content)
    
    def test_ajax_submission_skips_csrf_check(self):
        response = self.client.post('/safe-basic-form/', {
            'name': 'Test',
        }, HTTP_X_REQUESTED_WITH = 'XMLHttpRequest')
        self.assertEqual(response.content, 'Valid: Test')
    
    def test_ajax_submission_fails_check_if_ajax_skips_check_is_false(self):
        response = self.client.post('/safe-form-ajax-skips-false/', {
            'name': 'Test',
        }, HTTP_X_REQUESTED_WITH = 'XMLHttpRequest')
        self.assert_(CSRF_INVALID_MESSAGE in response.content)

class MultipleFormsTest(TestCase):
    urls = 'django_safeform.test_views'
    
    def test_two_safe_forms_does_not_result_in_duplicate_element_ids(self):
        response = self.client.get('/two-forms/')
        input_attrs = test_utils.extract_input_tag_attrs(response.content)
        ids = [d['id'] for d in input_attrs if 'id' in d]
        self.assertEqual(len(ids), len(set(ids)), 'Duplicate IDs in %s' % ids)

class CsrfFormTest(TestCase):
    urls = 'django_safeform.test_views'
    
    def test_csrf_form_can_protect_formsets(self):
        response = self.client.get('/safe-formset/')
        hiddens = test_utils.extract_input_tags(response.content)
        self.assert_('csrf_token' in hiddens)
        
        data = dict(hiddens)
        data['form-0-name'] = 'Simon'
        data['form-0-email'] = 'simon@example.com'
        data['form-1-name'] = 'Bob'
        data['form-1-email'] = 'bob@example.com'
        response2 = self.client.post('/safe-formset/', data)
        self.assertEqual(response2.content,
            'Valid: Simon [simon@example.com], Bob [bob@example.com]'
        )
        
        data['csrf_token'] = 'invalid-token'
        response3 = self.client.post('/safe-formset/', data)
        self.assert_(CSRF_INVALID_MESSAGE in response3.content)

class HandRolledFormsTest(TestCase):
    urls = 'django_safeform.test_views'
    
    def test_hand_rolled_forms_can_be_protected(self):
        response = self.client.get('/hand-rolled/')
        self.assert_(response.cookies.has_key('_csrf_cookie'))
        inputs = test_utils.extract_input_tags(response.content)
        self.assert_(inputs.has_key('csrf_token'))
        
        response2 = self.client.post('/hand-rolled/', {
            'csrf_token': 'bad',
            'name': 'Test',
        })
        self.assertEqual(response2.content, 'Invalid CSRF token')
        
        response3 = self.client.post('/hand-rolled/', {
            'csrf_token': inputs['csrf_token'],
            'name': 'Test',
        })
        self.assertEqual(response3.content, 'OK')

class IdentifierTest(TestCase):
    urls = 'django_safeform.test_views'
    
    def test_identifier_creates_tokens_that_only_work_with_one_form(self):
        # Get form token from a regular form
        response = self.client.get('/safe-basic-form/')
        token = test_utils.extract_input_tags(response.content)['csrf_token']
        self.assert_(token.startswith('default:'))
        response = self.client.post('/identifier-form/', {
            'name': 'Test',
            'csrf_token': token,
        })
        self.assert_(CSRF_INVALID_MESSAGE in response.content)
        
        # Now use the correct token
        response = self.client.get('/identifier-form/')
        token = test_utils.extract_input_tags(response.content)['csrf_token']
        self.assert_(token.startswith('identifier-form:'))
        response = self.client.post('/identifier-form/', {
            'name': 'Test 2',
            'csrf_token': token,
        })
        self.assertEqual(response.content, 'Valid: Test 2')

class ExpireAfterTest(TestCase):
    urls = 'django_safeform.test_views'
    
    def fetch_token(self, fake):
        @test_utils.fake_utcnow(fake)
        def inner():
            r = self.client.get('/safe-basic-form/')
            return test_utils.extract_input_tags(r.content)['csrf_token']
        return inner()
    
    def test_expires_after_argument_causes_token_to_expire(self):
        token = self.fetch_token(datetime.datetime(2009, 1, 1, 0, 0, 0))
        @test_utils.fake_utcnow(datetime.datetime(2009, 1, 1, 0, 0, 59))
        def submission_should_succeed():
            response = self.client.post('/expire-after-60-seconds/', {
                'name': 'Test',
                'csrf_token': token,
            })
            self.assertEqual(response.content, 'Valid: Test')
        submission_should_succeed()
        
        @test_utils.fake_utcnow(datetime.datetime(2009, 1, 1, 0, 1, 1))
        def submission_should_fail():
            response = self.client.post('/expire-after-60-seconds/', {
                'name': 'Test',
                'csrf_token': token,
            })
            self.assert_(CSRF_INVALID_MESSAGE in response.content)
        submission_should_fail()
    
    def test_expires_after_default_is_set_by_global_settings_py(self):
        token = self.fetch_token(datetime.datetime(2009, 1, 1, 0, 0, 0))
        # Test pretends to run 2 days later
        @test_utils.fake_utcnow(datetime.datetime(2009, 1, 3, 0, 0, 0))
        def inner():
            response = self.client.post('/safe-basic-form/', {
                'name': 'Test',
                'csrf_token': token,
            })
            self.assertEqual(response.content, 'Valid: Test')
            # Now monkey patch the settings
            from django.conf import settings
            settings.CSRF_TOKENS_EXPIRE_AFTER = 24 * 60 * 60
            response = self.client.post('/safe-basic-form/', {
                'name': 'Test',
                'csrf_token': token,
            })
            self.assert_(CSRF_INVALID_MESSAGE in response.content)
            settings.CSRF_TOKENS_EXPIRE_AFTER = None
        inner()



try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps  # Python 2.3, 2.4 fallback.

from django.conf import settings
from django import forms
from csrf_utils import new_csrf_token, validate_csrf_token

_ = lambda s: s

CSRF_INVALID_MESSAGE = _('Form session expired - please resubmit')

class HiddenInputNoId(forms.HiddenInput):
    def render(self, name, value, attrs=None):
        if attrs and 'id' in attrs:
            del attrs['id']
        return super(HiddenInputNoId, self).render(name, value, attrs)

not_set = object()

def SafeForm(form_class,
        identifier='default',
        invalid_message=CSRF_INVALID_MESSAGE,
        ajax_skips_check=True,
        expire_after=not_set
    ):
    class InnerSafeForm(form_class):
        def __init__(self, request, *args, **kwds):
            self.request = request
            if request.method == 'POST':
                super(InnerSafeForm, self).__init__(
                    request.POST, *args, **kwds
                )
            else:
                initial_data = kwds.get('initial', {})
                initial_data['csrf_token'] = new_csrf_token(
                    self.request,
                    identifier=identifier,
                )
                kwds['initial'] = initial_data
                super(InnerSafeForm, self).__init__(*args, **kwds)
            self.fields['csrf_token'] = forms.CharField(
                widget = HiddenInputNoId,
                required = False,
            )
        
        def clean(self):
            cleaned_data = super(InnerSafeForm, self).clean()
            token = cleaned_data.get('csrf_token', '')
            kwargs = dict(identifier=identifier)
            if expire_after is not not_set:
                kwargs['expire_after'] = expire_after
            if not token or not validate_csrf_token(
                    token, self.request, **kwargs
                ):
                # Our form is "in flight", and we want the user to be able to 
                # successfully resubmit it. This means we need to include a 
                # freshly generated CSRF token in the hidden form field for 
                # when the form is redisplayed with the validation error.
                if not (ajax_skips_check and self.request.is_ajax()):
                    self.data._mutable = True
                    self.data['csrf_token'] = new_csrf_token(self.request)
                    self.data._mutable = False
                    raise forms.ValidationError(invalid_message)
            return cleaned_data
    
    return wraps(form_class, updated=())(InnerSafeForm)

class CsrfForm(forms.Form):
    def __unicode__(self):
        # Default rendering should output the error list or nothing at all
        return self.as_p()
CsrfForm = SafeForm(CsrfForm)



from distutils.core import setup

setup(
    name = 'django-safeform',
    version = '1.0.1',
    url = 'http://github.com/simonw/django-safeform',
    author = 'Simon Willison',
    author_email = 'simon+safeform@simonwillison.net',
    description = 'CSRF protection for Django forms',
    long_description = open('README').read(),
    license = 'BSD',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    platforms = 'Any',
    packages = ['django_safeform']
)
